#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <windows.h>
#include <process.h>
#include <time.h>
#include <string>
using namespace std;

#define CYCLES 10
int gCount = 0;

#define MAX_SEM_COUNT 1
#define MAX_THREAD 3

std::string semaphoreName = "SEMAPHORE_TEST";
HANDLE hSemaphore;

typedef unsigned(__stdcall *THREAD_FUNC) (void *);
unsigned threadID[MAX_THREAD];
HANDLE hThread[MAX_THREAD];

int randomVal(int min, int max)
{
	static bool first = true;
	if (first)
	{
		srand(time(NULL)); //seeding for the first time only!
		first = false;
	}
	return min + rand() % (max - min);
}

void createSemaphore()
{
	hSemaphore = CreateSemaphore(NULL, 1, MAX_SEM_COUNT, semaphoreName.c_str()); 
	if (hSemaphore == NULL) {
		printf("CreateSemaphore(): error: %d.\n", GetLastError());
	}
	else {
		if (GetLastError() == ERROR_ALREADY_EXISTS)
			printf("CreateSemaphore(): opened existing %s semaphore.\n", semaphoreName.c_str());
		else
			printf("CreateSemaphore(): new %s semaphore successfully created.\n", semaphoreName.c_str());
	}
}

void openSemaphore()
{
	hSemaphore = OpenSemaphore(SEMAPHORE_ALL_ACCESS, FALSE, semaphoreName.c_str());
	if (hSemaphore == NULL) {
		printf("OpenSemaphore(): error: %d.\n", GetLastError());
	}
	else {
		printf("OpenSemaphore(): %s semaphore successfully opened.\n", semaphoreName.c_str());
	}
}

void closeSemaphore()
{
	if (hSemaphore)
		CloseHandle(hSemaphore);
}

unsigned lock()
{
	unsigned rc = WaitForSingleObject(hSemaphore, INFINITE);
	return rc;
}

unsigned unlock()
{
	unsigned rc = ReleaseSemaphore(hSemaphore, 1, NULL);
	return rc;
}

void* threadFunc(void* data)
{
	int threadIndex = (int)data;
	for (int i = 0; i < CYCLES; i++)
	{
		lock();
		printf("thread %d cycle i = %d global count = %d\n", threadIndex, i, gCount);
		gCount++;
		unlock();

		Sleep(randomVal(200, 500)); // window Sleep
	}

	return NULL;
}

int main(int argc, char** argv)
{
	// create semaphore
	createSemaphore();

	// create threads
	for (int i = 0; i < MAX_THREAD; i++)
	{
		hThread[i] = (HANDLE)_beginthreadex(NULL, 0, (THREAD_FUNC)threadFunc, (void *)i, 0, &threadID[i]);
		if (hThread[i] == 0) {
			printf("createThread(): error: %d\n", errno);
		}
		else {
			printf("createThread(): %d\n", hThread[i]);

		}
	}

	// main process loop
	for (int i = 0; i < 2*CYCLES; i++)
	{
		printf("main cycle i = %d\n", i);
		Sleep(randomVal(200,500)); // window Sleep
	}

	// wait for all threads to terminate
	WaitForMultipleObjects(MAX_THREAD, hThread, TRUE, INFINITE);
	// close thread
	for (int i = 0; i < MAX_THREAD; i++)
	{
		CloseHandle(hThread[i]); // _beginthreadex  close thread handle
	}
	// close semaphore
	closeSemaphore();
	return 0;
}